/*
# #########################################################
# This file contains the standard file I/O driver that is
# uses the DOSFS FAT file I/O routines.
#   
# Written by Dave Hein with contributions from other GCC
# development team members.
# Copyright (c) 2011 Parallax, Inc.
# MIT Licensed
# #########################################################
*/

#include <stdio.h>
#include <stdint.h>
#include <string.h>
#include <stdlib.h>
#include <sys/driver.h>
#include <sys/sd.h>
#include <compiler.h>
#include <errno.h>
#include <propeller.h>
#include "eefs.h"
#include "eefsdir.h"

#if defined(__PROPELLER_LMM__) || defined(__PROPELLER_CMM__)
#define FILE_BUFFER_SIZE 16     // Use a small file buffer for LMM
#else
#define FILE_BUFFER_SIZE 512    // Use a 512-byte file buffer for XMM/XMMC
#endif

// Global variables
//VOLINFO eefs_volinfo;
int eefs_mountflag = 0;
char eefs_currdir[MAX_PATH];
__attribute__((section(".hub"))) uint8_t eefs_scratch[512];

// Function prototypes
//extern void LoadSDDriver(uint32_t configwords[2]);
//int DFS_InitFileIO(void);
int eefs_stdio_errno(int errnum);
uint32_t eefs_mount(void);

// Define the the file prefix
static const char File_prefix[] = "";

// function called by fopen
static int File_fopen(FILE *fp, const char *fname1, const char *mode)
{
    int retval;
    FileHandleT *fileinfo;
    char fname[MAX_PATH];
    uint8_t eefs_mode;
    int fileinfosize = eefs_handlesize();
    char *fname2;

    if (!eefs_mountflag && eefs_mount())
    {
        errno = EIO;
        return -1;
    }

    eefs_resolve_path((char *)fname1, fname);
    fname2 = eefs_get_rootname(fname);

    if (!fname2)
    {
        errno = ENOENT;
        return -1;
    }

    fileinfo = malloc(fileinfosize + FILE_BUFFER_SIZE);

    if (!fileinfo)
    {
        errno = ENOMEM;
        return -1;
    }

    retval = eefs_hopen(fileinfo, fname2, mode[0]);

    if (retval)
    {
        errno = eefs_stdio_errno(retval);
        free(fileinfo);
        return -1;
    }

    fp->drvarg[0] = (int)fileinfo;

    // Set up file buffer in the FILE struct
    fp->_ptr = fp->_base = ((void *)fileinfo) + fileinfosize;
    fp->_bsiz = FILE_BUFFER_SIZE;
    fp->_flag |= _IOFREEBUF;

    return 0;
}

// function called by fclose
static int File_fclose(FILE *fp)
{
    FileHandleT *eefs_fp = (FileHandleT *)fp->drvarg[0];
    eefs_hclose(eefs_fp);
    free((void *)eefs_fp);
    return 0;
}

// function called by fread
static int File_read(FILE *fp, unsigned char *buf, int count)
{
    uint32_t successcount;
    FileHandleT *fileinfo = (FileHandleT *)fp->drvarg[0];

    successcount = eefs_hread(fileinfo, buf, count);
    if (successcount < 0)
    {
        errno = eefs_stdio_errno(successcount);
        successcount = 0;
    }
    return successcount;
}

// function called by fwrite
static int File_write(FILE *fp, unsigned char *buf, int count)
{
    uint32_t successcount;
    FileHandleT *fileinfo = (FileHandleT *)fp->drvarg[0];

    successcount = eefs_hwrite(fileinfo, buf, count);
    if (successcount < 0)
    {
        errno = eefs_stdio_errno(successcount);
        successcount = 0;
    }
    return successcount;
}

// function called by fseek
static int File_fseek(FILE *fp, long int offset, int origin)
{
    FileHandleT *eefs_fp = (FileHandleT *)fp->drvarg[0];
    if (origin == SEEK_CUR)
        offset += eefs_htell(eefs_fp);
    else if (origin == SEEK_END)
        offset += eefs_hget_filesize(eefs_fp);
    else if (origin != SEEK_SET)
    {
        errno = EIO;
        return -1;
    }
    if (offset < 0)
        offset = 0;
    eefs_hseek(eefs_fp, offset);
    return 0;
}

// function called by remove
static int File_remove(const char *fname)
{
    int errnum;
    char *fname2;
    int retval = 0;
    char path[MAX_PATH];

    eefs_resolve_path(fname, path);
    fname2 = eefs_get_rootname(path);

    if (!fname2)
    {
        errno = ENOENT;
        retval = -1;
    }
    else if ((errnum = eefs_hopen(0, fname2, 'd')))
    {
        errno = eefs_stdio_errno(errnum);
        retval = -1;
    }
    
    return retval;
}

// Define the driver list
_Driver _eeFileDriver =
{
    File_prefix,
    File_fopen,
    File_fclose,
    File_read,
    File_write,
    File_fseek,
    File_remove,
};

// Convert the dosfs errno to the stdio errno 
int eefs_stdio_errno(int errnum)
{
    return -errnum;
}

uint32_t eefs_mount(void)
{
    int retval = 0;

    if (eefs_mountflag) return 0;

    strcpy(eefs_currdir, "/");

    if (!eefs_mount_explicit(28, 29, 0x8000, 0x10000))
        eefs_mountflag = 1;
    else
        retval = -1;

    return retval;
}

static void movestr(char *ptr1, char *ptr2)
{
    while (*ptr2) *ptr1++ = *ptr2++;
    *ptr1 = 0;
}

char *eefs_get_rootname(char *path)
{
    char *rootname;
    int len = strlen(path);
    char *ptr = path + len - 1;
    int retval;

    while (len > 0)
    {
        if (*ptr == '/') break;
        ptr--;
        len--;
    }

    if (!len) return path;

    if (len == 1)
        retval = eefs_chdir("/");
    else
    {
        *ptr = 0;
        retval = eefs_chdir(path);
        *ptr = '/';
    }

    if (retval)
    {
        errno = -retval;
        return 0;
    }

    return ptr + 1;
}

void eefs_resolve_path(const char *fname, char *path)
{
    char *ptr;
    char *ptr1;

    if (!strcmp(fname, ".")) fname++;
    else if (!strncmp(fname, "./",2) || !strncmp(fname, ".\\",2)) fname += 2;

    if (fname[0] == '/' || fname[0] == '\\')
        strcpy(path, fname);
    else if (fname[0] == 0)
        strcpy(path, eefs_currdir);
    else
    {
        strcpy(path, eefs_currdir);
        int pathLen = strlen(path);
        if (path[pathLen-1] != '/' && path[pathLen-1] != '\\') strcat(path, "/");
        strcat(path, fname);
    }

    // Whack DOS paths to UNIX paths
    for (ptr = path;  *ptr;  ptr++)
        if (*ptr == '\\')
            *ptr = '/';

    // Process ..
    ptr = path;
    while (*ptr)
    {
        if (!strncmp(ptr, "/..", 3) && (ptr[3] == 0 || ptr[3] == '/'))
        {
            if (ptr == path)
            {
               movestr(ptr, ptr+3);
            }
            else
            {
                ptr1 = ptr - 1;
                while (ptr1 != path)
                {
                    if (*ptr1 == '/') break;
                    ptr1--;
                }
                movestr(ptr1, ptr+3);
                ptr = ptr1;
            }
        }
        else
        {
            ptr++;
            while (*ptr)
            {
               if (*ptr == '/') break;
               ptr++;
            }
        }
    }

    if (path[0] == 0) strcpy(path, "/");
}

/*
+--------------------------------------------------------------------
|  TERMS OF USE: MIT License
+--------------------------------------------------------------------
Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files
(the "Software"), to deal in the Software without restriction,
including without limitation the rights to use, copy, modify, merge,
publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so,
subject to the following conditions:

The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
+------------------------------------------------------------------
*/
